import 'dart:convert';
import 'dart:ui';

import 'package:crypto/crypto.dart';
import 'package:flutter/foundation.dart';
import 'package:intl/intl.dart';
import 'package:package_info_plus/package_info_plus.dart';
import 'package:phrase/src/cache/phrase_cache.dart';
import 'package:phrase/src/cache/phrase_web_cache.dart';
import 'package:synchronized/synchronized.dart';
import 'package:version/version.dart';

import 'cache/phrase_disk_cache.dart';
import 'phrase_api.dart';
import 'phrase_arb.dart';
import 'phrase_chef.dart';
import 'phrase_prefs.dart';

@visibleForTesting
const String kSdkVersion = "1.0.5"; //must match version in pubspec.yaml

final RegExp _kSemverRegexp = RegExp(
    r"^(0|[1-9]\d*)\.(0|[1-9]\d*)\.(0|[1-9]\d*)(-[a-zA-Z\d][-a-zA-Z.\d]*)?(\+[a-zA-Z\d][-a-zA-Z.\d]*)?$");
bool verifySemver(String ver) => _kSemverRegexp.hasMatch(ver);

typedef GetPlatformPackageInfoCallback = Future<PackageInfo> Function();

class PhraseImpl {
  PhraseImpl(
    this._distribution,
    this._environment,
    this._checkForUpdates,
    this._customAppVersion,
    this._host,
    this._defaultTimeout, {
    PhraseCache? localCache,
    PhrasePrefs? prefs,
    PhraseApi? api,
    PhraseChef? chef,
    GetPlatformPackageInfoCallback? getPlatformInfo,
  })  : _prefs = prefs ?? PhrasePrefs(),
        _chef = chef ?? PhraseChef(),
        _getPlatformInfo =
            getPlatformInfo ?? (() => PackageInfo.fromPlatform()) {
    _api = api ?? getPlatformApi();
    _localCache = localCache ?? (kIsWeb ? PhraseWebCache() : PhraseDiskCache());
  }

  final String _distribution;
  final String _environment;
  final bool _checkForUpdates;
  final String? _customAppVersion;
  late final PhraseCache _localCache;
  final PhrasePrefs _prefs;
  late final PhraseApi _api;
  final PhraseChef _chef;
  final PhraseHost _host;
  final Lock _lockApi = Lock();
  final Duration? _defaultTimeout;
  final GetPlatformPackageInfoCallback _getPlatformInfo;

  String? _currentLocale;
  AppResourceBundle? _appResourceBundle;

  @visibleForTesting
  String? get currentLocale => _currentLocale;
  @visibleForTesting
  AppResourceBundle? get appResourceBundle => _appResourceBundle;

  static String _hash(String s) => sha512.convert(utf8.encode(s)).toString();

  @visibleForTesting
  static String getLocaleHash(
    String distribution,
    String environment,
    String locale,
  ) =>
      _hash("$distribution-$environment-$locale");

  Future<void> onLocale(Locale locale) async {
    String localeCode = Intl.canonicalizedLocale(locale.toString());
    bool hasCache = await _loadCache(localeCode);
    if (_checkForUpdates) await loadFromApi(localeCode, hasCache);
  }

  Future<bool> _loadCache(String localeCode) async {
    String fileName = getLocaleHash(_distribution, _environment, localeCode);
    if (!(await _isCacheAppVersionValid())) {
      await _localCache.delete(fileName);
      await _prefs.resetVersion(fileName);
      await _prefs.resetAppVersionRange();
      await _prefs.resetLastUpdate();
      phraseLog("Deleted cache for `$localeCode`. App version not valid.");
      return false;
    }
    try {
      String? contents = await _localCache.read(fileName);
      if (contents != null) {
        _appResourceBundle = AppResourceBundle.parse(contents);
        _currentLocale = localeCode;
        phraseLog("Loaded cache for `$localeCode`");
        return true;
      } else {
        phraseLog("Cache for `$localeCode` not found");
      }
    } catch (e) {
      phraseLog("_loadCache failed with $e");
    }
    if (localeCode != _currentLocale) {
      _appResourceBundle = null;
    }
    return false;
  }

  Future<bool> _isCacheAppVersionValid() async {
    final appVersionStr = await _getAppVersion();
    if (appVersionStr == null) {
      return true;
    }
    final appVersion = Version.parse(appVersionStr);
    final (minAppVersion, maxAppVersion) = await _prefs.getAppVersionRange();
    final minVersionValid =
        minAppVersion == null || appVersion >= Version.parse(minAppVersion);
    final maxVersionValid =
        maxAppVersion == null || appVersion <= Version.parse(maxAppVersion);
    return minVersionValid && maxVersionValid;
  }

  Future<void> loadFromApi(
    String localeCode,
    bool hasCache, {
    Duration? timeout,
  }) =>
      _lockApi.synchronized(() async {
        String localeHash =
            getLocaleHash(_distribution, _environment, localeCode);
        String uuid = await _prefs.getUUID();
        String? sdkVersion = kSdkVersion;
        String? lastUpdate = await _prefs.getLastUpdate();
        String? currentVersion =
            hasCache ? await _prefs.getVersion(localeHash) : null;
        String? appVersion = _customAppVersion ?? await _getAppVersion();
        try {
          PhraseApiResult response = await _api.getTranslations(
            _distribution,
            _environment,
            _host,
            localeCode,
            uuid,
            sdkVersion,
            lastUpdate,
            currentVersion,
            appVersion,
            timeout ?? _defaultTimeout,
          );
          switch (response) {
            case PhraseApiResultSuccess(
                version: final version,
                arb: final arb,
                maxAppVersion: final maxAppVersion,
                minAppVersion: final minAppVersion,
              ):
              _appResourceBundle = AppResourceBundle.parse(arb);
              await _localCache.write(localeHash, arb);
              await _prefs.setLastUpdate();
              await _prefs.setVersion(localeHash, version);
              await _prefs.setAppVersionRange((
                minAppVersion,
                maxAppVersion,
              ));
              _currentLocale = localeCode;
              phraseLog("OTA update for `$localeCode`: OK");

            case PhraseApiResultNotModified():
              phraseLog("OTA update for `$localeCode`: NOT MODIFIED");

            case PhraseApiNotFound():
              _appResourceBundle = null;
              await _localCache.delete(localeHash);
              await _prefs.resetLastUpdate();
              await _prefs.resetVersion(localeHash);
              await _prefs.resetAppVersionRange();
              phraseLog("OTA update for `$localeCode`: NO RELEASES FOUND");
          }
        } catch (e) {
          phraseLog("OTA update for `$localeCode`: ERROR: $e");
          if (localeCode != _currentLocale) {
            _appResourceBundle = null;
          }
        }
      });

  Future<String?> _getAppVersion() async {
    try {
      String version = (await _getPlatformInfo()).version;
      if (!verifySemver(version)) {
        phraseLog(
            "WARNING: App version doesn't comply with semver specification: `$version`");
        return null;
      }
      return version;
    } catch (e) {
      return null;
    }
  }

  String? getText(
    String locale,
    String key, [
    Map<String, dynamic> args = const {},
  ]) {
    try {
      return _appResourceBundle == null
          ? null
          : _chef.cook(locale, _appResourceBundle!, key, args);
    } catch (e) {
      return null;
    }
  }
}

void phraseLog(String s) {
  if (kDebugMode) debugPrint("PHRASE OTA: $s");
}
