import 'package:flutter/material.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:uuid/uuid.dart';

typedef GetSharedPreferencesCallback = Future<SharedPreferences> Function();
typedef GenerateUuidCallback = String Function({Map<String, dynamic>? options});

typedef OtaReleaseAppVersionRange = (
  String? minAppVersion,
  String? maxAppVersion,
);

class PhrasePrefs {
  @visibleForTesting
  static const String kSdkInstallUuidKey = "phrase_uuid";
  @visibleForTesting
  static const String kLastUpdateKey = "phrase_last_update";
  static const String kMinAppVersionKey = "phrase_min_app_version";
  static const String kMaxAppVersionKey = "phrase_max_app_version";

  PhrasePrefs({
    GetSharedPreferencesCallback? getSharedPreferencesCallback,
    Uuid? uuid,
  })  : _getSharedPreferences =
            getSharedPreferencesCallback ?? SharedPreferences.getInstance,
        _uuidGenerator = uuid ?? const Uuid();

  final GetSharedPreferencesCallback _getSharedPreferences;
  final Uuid _uuidGenerator;

  Future<String> getUUID() async {
    String? uuid = await _get(kSdkInstallUuidKey);
    if (uuid != null) return uuid;
    uuid = _uuidGenerator.v4();
    await _put(kSdkInstallUuidKey, uuid);
    return uuid;
  }

  Future<String?> getLastUpdate() => _get(kLastUpdateKey);

  Future<void> setLastUpdate() => _put(kLastUpdateKey,
      (DateTime.now().millisecondsSinceEpoch / 1000).round().toString());

  Future<void> resetLastUpdate() => _remove(kLastUpdateKey);

  Future<String?> getVersion(String localeHash) =>
      _get(_keyTranslationVer(localeHash));

  Future<void> setVersion(String localeHash, String version) =>
      _put(_keyTranslationVer(localeHash), version);

  Future<void> resetVersion(String localeHash) =>
      _remove(_keyTranslationVer(localeHash));

  Future<OtaReleaseAppVersionRange> getAppVersionRange() async {
    final minAppVersion = await _get(kMinAppVersionKey);
    final maxAppVersion = await _get(kMaxAppVersionKey);
    return (minAppVersion, maxAppVersion);
  }

  Future<void> setAppVersionRange(
    OtaReleaseAppVersionRange appVersionRange,
  ) async {
    final (minAppVersion, maxAppVersion) = appVersionRange;
    if (minAppVersion != null) {
      await _put(kMinAppVersionKey, minAppVersion);
    } else {
      await _remove(kMinAppVersionKey);
    }
    if (maxAppVersion != null) {
      await _put(kMaxAppVersionKey, maxAppVersion);
    } else {
      await _remove(kMaxAppVersionKey);
    }
  }

  Future<void> resetAppVersionRange() async {
    await _remove(kMinAppVersionKey);
    await _remove(kMaxAppVersionKey);
  }

  String _keyTranslationVer(String localeHash) => "phrase_version_$localeHash";

  Future<String?> _get(String key) async {
    SharedPreferences prefs = await _getSharedPreferences();
    return prefs.getString(key);
  }

  Future<void> _put(String key, String value) async {
    SharedPreferences prefs = await _getSharedPreferences();
    await prefs.setString(key, value);
  }

  Future<void> _remove(String key) async {
    SharedPreferences prefs = await _getSharedPreferences();
    await prefs.remove(key);
  }
}
