import 'package:flutter/material.dart';
import 'package:mock_data/mock_data.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:package_info_plus/package_info_plus.dart';
import 'package:phrase/src/cache/phrase_disk_cache.dart';
import 'package:phrase/src/phrase_api.dart';
import 'package:phrase/src/phrase_arb.dart';
import 'package:phrase/src/phrase_chef.dart';
import 'package:phrase/src/phrase_impl.dart';
import 'package:phrase/src/phrase_prefs.dart';
import 'package:test/test.dart';

import '../_helper/arb.dart';
import 'phrase_impl_test.mocks.dart';

@GenerateMocks([
  PhraseDiskCache,
  PhrasePrefs,
  PhraseApi,
  PhraseChef,
])
void main() {
  group('PhraseImpl', () {
    final diskCacheMock = MockPhraseDiskCache();
    final prefsMock = MockPhrasePrefs();
    final apiMock = MockPhraseApi();
    final chefMock = MockPhraseChef();

    final distId = mockUUID();
    final envId = mockUUID();
    const timeout = Duration(seconds: 20);

    setUp(() {
      reset(diskCacheMock);
      reset(prefsMock);
      reset(apiMock);
      reset(chefMock);
    });

    PhraseImpl createPhraseImpl({
      GetPlatformPackageInfoCallback? getPlatformPackageInfo,
    }) =>
        PhraseImpl(
          distId,
          envId,
          true,
          null,
          PhraseHost.eu,
          timeout,
          localCache: diskCacheMock,
          prefs: prefsMock,
          api: apiMock,
          chef: chefMock,
          getPlatformInfo: getPlatformPackageInfo,
        );

    group('loadFromApi', () {
      final uuid = mockUUID();
      const locale = 'de';
      const version = '1';
      final lastUpdate = DateTime.now().toIso8601String();
      final arb = loadArb('test.arb');
      final localeHash = PhraseImpl.getLocaleHash(distId, envId, locale);

      setUp(() {
        when(prefsMock.getUUID()).thenAnswer((realInvocation) async => uuid);
        when(prefsMock.getVersion(any))
            .thenAnswer((realInvocation) async => version);
        when(prefsMock.getLastUpdate())
            .thenAnswer((realInvocation) async => lastUpdate);
      });

      test('loads modified translations', () async {
        final apiResult = PhraseApiResult.success(version, arb);
        provideDummy<PhraseApiResult>(apiResult);

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);
        final expectedArb = AppResourceBundle.parse(arb);

        final phraseImpl = createPhraseImpl();

        await phraseImpl.loadFromApi(locale, true);

        expect(
          phraseImpl.appResourceBundle?.resources,
          expectedArb.resources,
        );
        expect(
          phraseImpl.appResourceBundle?.resourceIds,
          expectedArb.resourceIds,
        );
        expect(phraseImpl.currentLocale, locale);

        verify(diskCacheMock.write(localeHash, arb));
        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(prefsMock.setLastUpdate());
        verify(prefsMock.setVersion(localeHash, version));
        verify(prefsMock.setAppVersionRange((null, null)));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
          timeout,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('translations not modified', () async {
        final apiResult = PhraseApiResult.notModified();
        provideDummy<PhraseApiResult>(apiResult);

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        final phraseImpl = createPhraseImpl();

        await phraseImpl.loadFromApi(locale, true);

        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
          timeout,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('loads from us host when set', () async {
        final phraseImpl = PhraseImpl(
          distId,
          envId,
          true,
          null,
          PhraseHost.us,
          null,
          localCache: diskCacheMock,
          prefs: prefsMock,
          api: apiMock,
          chef: chefMock,
        );
        final apiResult = PhraseApiResult.success(version, arb);
        provideDummy<PhraseApiResult>(apiResult);

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        await phraseImpl.loadFromApi(locale, true);

        verify(diskCacheMock.write(localeHash, arb));
        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(prefsMock.setLastUpdate());
        verify(prefsMock.setVersion(localeHash, version));
        verify(prefsMock.setAppVersionRange((null, null)));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.us,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
          null,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('deletes local cache on 404 response', () async {
        final apiResult = PhraseApiResult.notFound();
        provideDummy<PhraseApiResult>(apiResult);

        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);

        final phraseImpl = createPhraseImpl();

        await phraseImpl.loadFromApi(locale, true);

        verifyInOrder([
          prefsMock.getUUID(),
          prefsMock.getLastUpdate(),
          prefsMock.getVersion(localeHash),
          apiMock.getTranslations(
            distId,
            envId,
            PhraseHost.eu,
            locale,
            uuid,
            kSdkVersion,
            lastUpdate,
            version,
            null,
            timeout,
          ),
          diskCacheMock.delete(localeHash),
          prefsMock.resetLastUpdate(),
          prefsMock.resetVersion(localeHash),
          prefsMock.resetAppVersionRange(),
        ]);

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });

      test('fails to load translation', () async {
        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any, any))
            .thenThrow(Exception());

        final phraseImpl = createPhraseImpl();

        await phraseImpl.loadFromApi(locale, true);

        expect(phraseImpl.appResourceBundle, null);
        expect(phraseImpl.currentLocale, null);

        verify(prefsMock.getUUID());
        verify(prefsMock.getLastUpdate());
        verify(prefsMock.getVersion(localeHash));
        verify(apiMock.getTranslations(
          distId,
          envId,
          PhraseHost.eu,
          locale,
          uuid,
          kSdkVersion,
          lastUpdate,
          version,
          null,
          timeout,
        ));

        verifyNoMoreInteractions(diskCacheMock);
        verifyNoMoreInteractions(prefsMock);
        verifyNoMoreInteractions(apiMock);
        verifyZeroInteractions(chefMock);
      });
    });

    group('getText', () {
      final phraseImpl = PhraseImpl(
        distId,
        envId,
        true,
        null,
        PhraseHost.eu,
        null,
        localCache: diskCacheMock,
        prefs: prefsMock,
        api: apiMock,
        chef: chefMock,
      );

      final uuid = mockUUID();
      const locale = 'de';
      const version = '1';
      final lastUpdate = DateTime.now().toIso8601String();
      final arbFileContents = loadArb('test.arb');

      setUp(() async {
        when(prefsMock.getUUID()).thenAnswer((realInvocation) async => uuid);
        when(prefsMock.getVersion(any))
            .thenAnswer((realInvocation) async => version);
        when(prefsMock.getLastUpdate())
            .thenAnswer((realInvocation) async => lastUpdate);

        final apiResult = PhraseApiResult.success(version, arbFileContents);
        provideDummy<PhraseApiResult>(apiResult);
        when(apiMock.getTranslations(
                any, any, any, any, any, any, any, any, any, any))
            .thenAnswer((realInvocation) async => apiResult);
        await phraseImpl.loadFromApi(locale, true);

        reset(diskCacheMock);
        reset(prefsMock);
        reset(apiMock);
        reset(chefMock);
      });

      test('success', () async {
        const key = 'text';
        when(chefMock.cook(any, any, any)).thenReturn('expected');

        final result = phraseImpl.getText(locale, key);

        expect(result, 'expected');

        verify(chefMock.cook(locale, any, key));
        verifyNoMoreInteractions(chefMock);
        verifyZeroInteractions(diskCacheMock);
        verifyZeroInteractions(prefsMock);
        verifyZeroInteractions(apiMock);
      });

      test('translation not found', () {
        const key = 'text';
        when(chefMock.cook(any, any, any)).thenReturn(null);
        final result = phraseImpl.getText(locale, key);

        expect(result, null);

        verify(chefMock.cook(locale, any, key));
        verifyNoMoreInteractions(chefMock);
        verifyZeroInteractions(diskCacheMock);
        verifyZeroInteractions(prefsMock);
        verifyZeroInteractions(apiMock);
      });
    });

    group('onLocale', () {
      const localeCode = 'de';
      const locale = Locale(localeCode);
      final fileName = PhraseImpl.getLocaleHash(distId, envId, localeCode);
      final arb = loadArb('test.arb');
      final packageInfo = PackageInfo(
        appName: 'Test App',
        packageName: 'com.example.testapp',
        version: '1.0.0',
        buildNumber: '1',
      );

      setUp(() {
        when(diskCacheMock.read(any)).thenAnswer((realInvocation) async => arb);
        when(prefsMock.getUUID())
            .thenAnswer((realInvocation) async => mockUUID());
        when(prefsMock.getLastUpdate())
            .thenAnswer((realInvocation) async => null);
        when(prefsMock.getAppVersionRange())
            .thenAnswer((_) async => (null, null));
      });

      group('loads from cache', () {
        late PhraseImpl phraseImpl;
        setUp(() {
          phraseImpl = PhraseImpl(
            distId,
            envId,
            false,
            null,
            PhraseHost.eu,
            null,
            localCache: diskCacheMock,
            prefs: prefsMock,
            api: apiMock,
            chef: chefMock,
            getPlatformInfo: () async => packageInfo,
          );
        });

        test('on success update resource bundle and current locale', () async {
          final expectedArb = AppResourceBundle.parse(arb);

          await phraseImpl.onLocale(locale);

          expect(
            phraseImpl.appResourceBundle?.resources,
            expectedArb.resources,
          );
          expect(
            phraseImpl.appResourceBundle?.resourceIds,
            expectedArb.resourceIds,
          );
          expect(phraseImpl.currentLocale, localeCode);

          verify(diskCacheMock.read(fileName));
          verifyNoMoreInteractions(diskCacheMock);
          verifyZeroInteractions(apiMock);
          verifyZeroInteractions(chefMock);
        });

        test('on failure sets resource bundle and current locale to null',
            () async {
          when(diskCacheMock.read(any)).thenThrow(Exception());
          await phraseImpl.onLocale(locale);
          expect(phraseImpl.appResourceBundle, null);
          expect(phraseImpl.currentLocale, null);
        });
      });

      test('loads local cache when app version is in of range', () async {
        when(prefsMock.getAppVersionRange())
            .thenAnswer((_) async => ('2.0.0', '3.0.0'));
        when(prefsMock.getVersion(any))
            .thenAnswer((realInvocation) async => null);
        final phraseImpl = createPhraseImpl(
            getPlatformPackageInfo: () async => PackageInfo(
                  appName: 'Test App',
                  packageName: 'com.example.testapp',
                  version: '2.5.0',
                  buildNumber: '1',
                ));

        await phraseImpl.onLocale(locale);

        verify(diskCacheMock.read(fileName));
        verifyNever(diskCacheMock.delete(any));
      });

      test('deletes local cache when app version is out of range', () async {
        when(prefsMock.getAppVersionRange())
            .thenAnswer((_) async => ('2.0.0', '3.0.0'));
        final phraseImpl = createPhraseImpl(
            getPlatformPackageInfo: () async => PackageInfo(
                  appName: 'Test App',
                  packageName: 'com.example.testapp',
                  version: '1.0.0',
                  buildNumber: '1',
                ));

        await phraseImpl.onLocale(locale);

        verify(diskCacheMock.delete(fileName));
        verify(prefsMock.resetVersion(fileName));
        verify(prefsMock.resetAppVersionRange());
        verify(prefsMock.resetLastUpdate());
        verifyNever(diskCacheMock.read(any));
      });

      group('loads translations from api', () {
        final uuid = mockUUID();
        const version = '1';
        final lastUpdate = DateTime.now().toIso8601String();
        final apiResult = PhraseApiResult.success(version, arb);
        late PhraseImpl phraseImpl;

        setUp(() {
          phraseImpl = PhraseImpl(
            distId,
            envId,
            true,
            null,
            PhraseHost.eu,
            null,
            localCache: diskCacheMock,
            prefs: prefsMock,
            api: apiMock,
            chef: chefMock,
          );

          when(prefsMock.getUUID()).thenAnswer((realInvocation) async => uuid);
          when(prefsMock.getVersion(any))
              .thenAnswer((realInvocation) async => version);
          when(prefsMock.getLastUpdate())
              .thenAnswer((realInvocation) async => lastUpdate);
          when(apiMock.getTranslations(
                  any, any, any, any, any, any, any, any, any, any))
              .thenAnswer((realInvocation) async => apiResult);
        });

        test('on success update resource bundle and current locale', () async {
          final expectedArb = AppResourceBundle.parse(arb);

          await phraseImpl.onLocale(locale);

          expect(
            phraseImpl.appResourceBundle?.resources,
            expectedArb.resources,
          );
          expect(
            phraseImpl.appResourceBundle?.resourceIds,
            expectedArb.resourceIds,
          );
          expect(phraseImpl.currentLocale, localeCode);

          verify(diskCacheMock.read(fileName));
          verify(apiMock.getTranslations(
            distId,
            envId,
            PhraseHost.eu,
            localeCode,
            uuid,
            kSdkVersion,
            lastUpdate,
            version,
            null,
            null,
          ));
        });

        test('on load cache failure keep app resource bundle', () async {
          when(diskCacheMock.read(any)).thenThrow(Exception());
          final expectedArb = AppResourceBundle.parse(arb);

          await phraseImpl.onLocale(locale);

          expect(
            phraseImpl.appResourceBundle?.resources,
            expectedArb.resources,
          );
          expect(
            phraseImpl.appResourceBundle?.resourceIds,
            expectedArb.resourceIds,
          );
          expect(phraseImpl.currentLocale, localeCode);
        });
      });
    });
  });
}
